@@ -4,7 +4,7 @@ from glob import glob
 from os.path import basename, dirname, join
 from typing import Optional
 
-from pandas import StringDtype
+from pandas import DataFrame, read_csv
 from xarray import DataArray
 
 from ._allele import Allele
@@ -24,12 +24,13 @@ def read_plink(file_prefix, verbose=True):
         The function :func:`pandas_plink.read_plink1_bin` provides an alternative
         interface to read the same files.
 
-    The genotype values can be either ``0``, ``1``, ``2``, or ``NaN``:
+    The genotype values can be either :const:`0`, :const:`1`, :const:`2`, or
+    :data:`math.nan`:
 
-    - ``0`` Homozygous having the first allele (given by coordinate ``a0``)
-    - ``1`` Heterozygous
-    - ``2`` Homozygous having the second allele (given by coordinate ``a1``)
-    - ``NaN`` Missing genotype
+    - :const:`0` Homozygous having the first allele (given by coordinate **a0**)
+    - :const:`1` Heterozygous
+    - :const:`2` Homozygous having the second allele (given by coordinate **a1**)
+    - :data:`math.nan` Missing genotype
 
     Examples
     --------
@@ -42,7 +43,8 @@ def read_plink(file_prefix, verbose=True):
         >>> from os.path import join
         >>> from pandas_plink import read_plink
         >>> from pandas_plink import get_data_folder
-        >>> (bim, fam, bed) = read_plink(join(get_data_folder(), "chr*.bed"), verbose=False)
+        >>> (bim, fam, bed) = read_plink(join(get_data_folder(), "chr*.bed"),
+        ...                              verbose=False)
         >>> print(bim.head())
           chrom        snp       cm     pos a0 a1  i
         0    11  316849996     0.00  157439  C  T  0
@@ -72,7 +74,7 @@ def read_plink(file_prefix, verbose=True):
         Path prefix to the set of PLINK files. It supports loading many BED files at
         once using globstrings wildcard.
     verbose : bool
-        ``True`` for progress information; ``False`` otherwise.
+        :const:`True` for progress information; :const:`False` otherwise.
 
     Returns
     -------
@@ -119,7 +121,7 @@ def read_plink(file_prefix, verbose=True):
     ref = Allele.a1
     bed = _read_file(
         fn,
-        lambda f: _read_bed(f["bed"], nsamples, nmarkers[f["bed"]], ref, Chunk()),
+        lambda f: _read_bed(f["bed"], nsamples, nmarkers[f["bed"]], ref, Chunk()).T,
         pbar,
     )
 
@@ -172,15 +174,14 @@ def read_plink1_bin(
         >>> print(G)
         <xarray.DataArray 'genotype' (sample: 14, variant: 1252)>
         dask.array<concatenate, shape=(14, 1252), dtype=float32, chunksize=(14, 779), chunktype=numpy.ndarray>
-        Coordinates:
+        Coordinates: (12/14)
           * sample   (sample) object 'B001' 'B002' 'B003' ... 'B012' 'B013' 'B014'
-          * variant  (variant) object '11_316849996' '11_316874359' ... '12_373081507'
+          * variant  (variant) <U11 'variant0' 'variant1' ... 'variant1251'
             fid      (sample) object 'B001' 'B002' 'B003' ... 'B012' 'B013' 'B014'
             iid      (sample) object 'B001' 'B002' 'B003' ... 'B012' 'B013' 'B014'
             father   (sample) object '0' '0' '0' '0' '0' '0' ... '0' '0' '0' '0' '0' '0'
             mother   (sample) object '0' '0' '0' '0' '0' '0' ... '0' '0' '0' '0' '0' '0'
-            gender   (sample) object '0' '0' '0' '0' '0' '0' ... '0' '0' '0' '0' '0' '0'
-            trait    (sample) object '-9' '-9' '-9' '-9' '-9' ... '-9' '-9' '-9' '-9'
+            ...       ...
             chrom    (variant) object '11' '11' '11' '11' '11' ... '12' '12' '12' '12'
             snp      (variant) object '316849996' '316874359' ... '373081507'
             cm       (variant) float64 0.0 0.0 0.0 0.0 0.0 0.0 ... 0.0 0.0 0.0 0.0 0.0
@@ -198,15 +199,14 @@ def read_plink1_bin(
         >>> print(G)
         <xarray.DataArray 'genotype' (sample: 14, variant: 779)>
         dask.array<where, shape=(14, 779), dtype=float32, chunksize=(14, 779), chunktype=numpy.ndarray>
-        Coordinates:
+        Coordinates: (12/14)
           * sample   (sample) object 'B001' 'B002' 'B003' ... 'B012' 'B013' 'B014'
-          * variant  (variant) object '11_316849996' '11_316874359' ... '11_345698259'
+          * variant  (variant) <U11 'variant0' 'variant1' ... 'variant777' 'variant778'
             fid      (sample) object 'B001' 'B002' 'B003' ... 'B012' 'B013' 'B014'
             iid      (sample) object 'B001' 'B002' 'B003' ... 'B012' 'B013' 'B014'
             father   (sample) object '0' '0' '0' '0' '0' '0' ... '0' '0' '0' '0' '0' '0'
             mother   (sample) object '0' '0' '0' '0' '0' '0' ... '0' '0' '0' '0' '0' '0'
-            gender   (sample) object '0' '0' '0' '0' '0' '0' ... '0' '0' '0' '0' '0' '0'
-            trait    (sample) object '-9' '-9' '-9' '-9' '-9' ... '-9' '-9' '-9' '-9'
+            ...       ...
             chrom    (variant) object '11' '11' '11' '11' '11' ... '11' '11' '11' '11'
             snp      (variant) object '316849996' '316874359' ... '345698259'
             cm       (variant) float64 0.0 0.0 0.0 0.0 0.0 0.0 ... 0.0 0.0 0.0 0.0 0.0
@@ -216,11 +216,11 @@ def read_plink1_bin(
         >>> print(G.shape)
         (14, 779)
 
-    Lets now print the genotype value of the sample `B003` for variant `11_316874359`:
+    Lets now print the genotype value of the sample `B003` for variant `variant5`:
 
     .. doctest::
 
-        >>> print(G.sel(sample="B003", variant="11_316874359").values)
+        >>> print(G.sel(sample="B003", variant="variant5").values)
         0.0
 
     The special matrix we return is of type :class:`xarray.DataArray`. More information
@@ -234,15 +234,15 @@ def read_plink1_bin(
         BED files.
     bim
         Path to a BIM file. It can contain shell-style wildcards to indicate multiple
-        BIM files. It defaults to ``None``, in which case it will try to be inferred.
+        BIM files. It defaults to :const:`None`, in which case it will try to be inferred.
     fam
-        Path to a FAM file. It defaults to ``None``, in which case it will try to be
+        Path to a FAM file. It defaults to :const:`None`, in which case it will try to be
         inferred.
     verbose
-        ``True`` for progress information; ``False`` otherwise.
+        :const:`True` for progress information; :const:`False` otherwise.
     ref
         Reference allele. Specify which allele the dosage matrix will count. It can
-        be either ``"a1"`` (default) or ``"a0"``.
+        be either :const:`"a1"` (default) or :const:`"a0"`.
     chunk
         Data chunk specification. Useful to adjust the trade-off between computational
         overhead and IO usage. See :class:`pandas_plink.Chunk`.
@@ -291,16 +291,15 @@ def read_plink1_bin(
     nfiles = len(bed_files) + len(bim_files) + 1
     pbar = tqdm(desc="Mapping files", total=nfiles, disable=not verbose)
 
-    bims = _read_file(bim_files, lambda f: _read_bim(f), pbar)
+    bims = _read_file(bim_files, lambda f: _read_bim_noi(f), pbar)
     nmarkers = {bed_files[i]: b.shape[0] for i, b in enumerate(bims)}
     bim_df = pd.concat(bims, axis=0, ignore_index=True)
-    del bim_df["i"]
-    fam_df = _read_file(fam_files, lambda f: _read_fam(f), pbar)[0]
-    del fam_df["i"]
+    fam_df = _read_file(fam_files, lambda f: _read_fam_noi(f), pbar)[0]
 
     nsamples = fam_df.shape[0]
     sample_ids = fam_df["iid"]
-    variant_ids = bim_df[["chrom", "snp"]].agg("_".join, axis=1)
+    nvariants = bim_df.shape[0]
+    variant_ids = [f"variant{i}" for i in range(nvariants)]
 
     if ref == "a1":
         ref_al = Allele.a1
@@ -310,13 +309,13 @@ def read_plink1_bin(
         raise ValueError("Unknown reference allele.")
 
     G = _read_file(
-        bed_files, lambda f: _read_bed(f, nsamples, nmarkers[f], ref_al, chunk).T, pbar
+        bed_files, lambda f: _read_bed(f, nsamples, nmarkers[f], ref_al, chunk), pbar
     )
     G = da.concatenate(G, axis=1)
 
     G = DataArray(G, dims=["sample", "variant"], coords=[sample_ids, variant_ids])
     sample = {c: ("sample", fam_df[c]) for c in fam_df.columns}
-    variant = {c: ("variant", bim_df[c]) for c in bim_df.columns}
+    variant = {c: ("variant", bim_df[c]) for c in iter(bim_df.columns)}
     G = G.assign_coords(**sample)
     G = G.assign_coords(**variant)
     G.name = "genotype"
@@ -334,10 +333,9 @@ def _read_file(fn, read_func, pbar):
     return data
 
 
-def _read_csv(fn, header):
-    from pandas import read_csv
+def _read_csv(fn, header) -> DataFrame:
 
-    return read_csv(
+    df = read_csv(
         fn,
         delim_whitespace=True,
         header=None,
@@ -345,61 +343,77 @@ def _read_csv(fn, header):
         dtype=header,
         compression=None,
         engine="c",
+        iterator=False,
     )
+    assert isinstance(df, DataFrame)
+    return df
 
 
 def _read_bim(fn):
-    from numpy import float64, int32
+    df = _read_bim_noi(fn)
+    df["i"] = range(df.shape[0])
+    return df
+
+
+def _read_bim_noi(fn):
+    from ._type import bim
 
     header = odict(
         [
-            ("chrom", StringDtype()),
-            ("snp", StringDtype()),
-            ("cm", float64),
-            ("pos", int32),
-            ("a0", StringDtype()),
-            ("a1", StringDtype()),
+            ("chrom", bim["chrom"]),
+            ("snp", bim["snp"]),
+            ("cm", bim["cm"]),
+            ("pos", bim["pos"]),
+            ("a0", bim["a0"]),
+            ("a1", bim["a1"]),
         ]
     )
-    df = _read_csv(fn, header)
-
-    df["i"] = range(df.shape[0])
-    return df
+    return _read_csv(fn, header)
 
 
 def _read_fam(fn):
+    df = _read_fam_noi(fn)
+    df["i"] = range(df.shape[0])
+    return df
+
+
+def _read_fam_noi(fn):
+    from ._type import fam
+
     header = odict(
         [
-            ("fid", StringDtype()),
-            ("iid", StringDtype()),
-            ("father", StringDtype()),
-            ("mother", StringDtype()),
-            ("gender", StringDtype()),
-            ("trait", StringDtype()),
+            ("fid", fam["fid"]),
+            ("iid", fam["iid"]),
+            ("father", fam["father"]),
+            ("mother", fam["mother"]),
+            ("gender", fam["gender"]),
+            ("trait", fam["trait"]),
         ]
     )
 
-    df = _read_csv(fn, header)
-
-    df["i"] = range(df.shape[0])
-    return df
+    return _read_csv(fn, header)
 
 
 def _read_bed(fn, nsamples, nvariants, ref: Allele, chunk: Chunk):
     _check_bed_header(fn)
     major = _major_order(fn)
 
-    ncols = nvariants
-    nrows = nsamples
-    row_chunk = nrows if chunk.nsamples is None else min(nrows, chunk.nsamples)
-    col_chunk = ncols if chunk.nvariants is None else min(ncols, chunk.nvariants)
+    # Assume major == "variant".
+    nrows = nvariants
+    ncols = nsamples
+    row_chunk = nrows if chunk.nvariants is None else min(nrows, chunk.nvariants)
+    col_chunk = ncols if chunk.nsamples is None else min(ncols, chunk.nsamples)
 
-    if major == "variant":
+    if major == "sample":
         nrows, ncols = ncols, nrows
         row_chunk, col_chunk = col_chunk, row_chunk
 
+    max_npartitions = 16_384
+    row_chunk = max(nrows // max_npartitions, row_chunk)
+    col_chunk = max(ncols // max_npartitions, col_chunk)
+
     G = read_bed(fn, nrows, ncols, row_chunk, col_chunk, ref)
-    if major == "sample":
+    if major == "variant":
         G = G.T
 
     return G
@@ -416,6 +430,13 @@ def _check_bed_header(fn):
 
 
 def _major_order(fn):
+    """
+    Major order.
+
+    Variant-major lists all samples for first variant, all samples for second
+    variant, and so on. Sample-major lists all variants for first sample, all
+    variants for second sample, and so on.
+    """
     with open(fn, "rb") as f:
         f.seek(2)
         arr = f.read(1)
